/**
 * Copyright 2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 */
package com.sunlabs.archsim.fleet.parser.program;

import java.io.FileNotFoundException;
import java.util.Iterator;
import java.util.LinkedList;

import com.sunlabs.archsim.fleet.codeStorage.CodeBag;
import com.sunlabs.archsim.fleet.codeStorage.CodeStorage;
import com.sunlabs.archsim.fleet.common.Data;
import com.sunlabs.archsim.fleet.common.Payload;
import com.sunlabs.archsim.fleet.nameTable.SourceDestinationNameTable;

import junit.framework.TestCase;
import junit.textui.TestRunner;

/**
 * Test for fleet program parser/compiler.
 * 
 * @author ib27688
 *
 */
public class ProgramParserTest extends TestCase {

	// path for input files...
	private String path= "src/" +
		ProgramParserTest.class.getPackage().getName().replace('.','/') + "/";
	
	protected void setUp() throws Exception {
		super.setUp();
		SourceDestinationNameTable.getInstance().clear();
		CodeStorage.getInstance().clear();
	}

	protected void tearDown() throws Exception {
		super.tearDown();
	}

	public static void main(String[] args) {
		// run the tests
		TestRunner.run(ProgramParserTest.class);
		// ProgramParserTest ppt= new ProgramParserTest();
		// ppt.testGoodProgram();
	}
	
	public void testNoFile() {
		SourceDestinationNameTable.getInstance().clear();
		CodeStorage.getInstance().clear();
		
		ProgramParser programParser= new ProgramParser();
		boolean exceptionOK= false;
		try {
			programParser.parse("NoSuchFile");
		}
		catch (Exception e) {
			exceptionOK= (e instanceof FileNotFoundException);
		}
		assert(exceptionOK);
	}	
	
	public void testGoodProgram() {
		SourceDestinationNameTable.getInstance().clear();
		CodeStorage.getInstance().clear();
		
		ProgramParser parser= new ProgramParser();
		boolean noException= true;
		try {
			parser.parse(path+"goodProgram.fleet");		}
		catch (Exception e) {
			noException= false;
			System.out.println("Exception message: " + e.getMessage());
		}
		// System.out.println(SourceDestinationNameTable.getInstance().toString());
		// System.out.println(CodeStorage.getInstance().toSting());
		assertTrue(noException);
				
		// now see whether we got aliases right
		try {
			SourceDestinationNameTable map= SourceDestinationNameTable.getInstance();
			assertEquals(0, map.getSourceAddress("s0"));
			assertEquals(1, map.getSourceAddress("s1"));
			assertEquals(2, map.getSourceAddress("s2"));
			assertEquals(3, map.getSourceAddress("s3"));
			assertEquals(4, map.getSourceAddress("s4"));
			assertEquals(5, map.getSourceAddress("s5"));
			assertEquals(6, map.getSourceAddress("s6"));
			assertEquals(7, map.getSourceAddress("s7"));
			
			assertEquals(0, map.getDestinationAddress("d0"));
			assertEquals(1, map.getDestinationAddress("d1"));
			assertEquals(2, map.getDestinationAddress("d2"));
			assertEquals(3, map.getDestinationAddress("d3"));
			assertEquals(4, map.getDestinationAddress("d4"));
			assertEquals(5, map.getDestinationAddress("d5"));
			assertEquals(6, map.getDestinationAddress("d6"));
			assertEquals(7, map.getDestinationAddress("d7"));
		}
		catch (Exception e) {
			fail();
		}
		
		// check code bags
		CodeStorage codeStorage= CodeStorage.getInstance();

		// reference list for startup bags
		LinkedList<CodeBag> startupRefList= new LinkedList<CodeBag>();

		
		// check out bag "ONE"
		assertTrue(codeStorage.containsBag("ONE"));
		CodeBag bag= codeStorage.getBag("ONE");

		Payload p= new Payload();
		p.source= 0;
		p.destination= 0;
		p.data= null;
		p.repeats= 1;
		p.type= Payload.PayloadType.INSTRUCTION;
		// System.out.println("\n\nPayload= " + p + "\n\n");
		assertTrue(bag.instructions().contains(p));
		
		p= new Payload();
		p.source= 1;
		p.destination= 1;
		p.data= null;
		p.repeats= 1;
		p.type= Payload.PayloadType.INSTRUCTION;
		// System.out.println("\n\nPayload= " + p + "\n\n");
		assertTrue(bag.instructions().contains(p));

		// now bag "TWO"
		assertTrue(codeStorage.containsBag("TWO"));
		bag= codeStorage.getBag("TWO");
		// "TWO" is a startup bag
		startupRefList.add(bag);
		
		p= new Payload();
		p.source= 6;
		p.destination= 3;
		p.data= null;
		p.repeats= 1;
		p.type= Payload.PayloadType.INSTRUCTION;
		// System.out.println("\n\nPayload= " + p + "\n\n");
		assertTrue(bag.instructions().contains(p));
		
		p= new Payload();
		p.source= 4;
		p.destination= 7;
		p.data= null;
		p.repeats= 1;
		p.type= Payload.PayloadType.INSTRUCTION;
		// System.out.println("\n\nPayload= " + p + "\n\n");
		assertTrue(bag.instructions().contains(p));
		
		p= new Payload();
		p.source= 5;
		p.destination= 6;
		p.data= null;
		p.repeats= 1;
		p.type= Payload.PayloadType.INSTRUCTION;
		// System.out.println("\n\nPayload= " + p + "\n\n");
		assertTrue(bag.instructions().contains(p));

		// now bag "THREE"
		assertTrue(codeStorage.containsBag("THREE"));
		bag= codeStorage.getBag("THREE");
		
		p= new Payload();
		p.source= -1;
		p.destination= 0;
		p.data= new Data(0);
		p.repeats= 1;
		p.type= Payload.PayloadType.LITERAL;
		// System.out.println("\n\nPayload= " + p + "\n\n");
		assertTrue(bag.instructions().contains(p));

		p= new Payload();
		p.source= -1;
		p.destination= 7;
		p.data= new Data("fleet");
		p.repeats= 1;
		p.type= Payload.PayloadType.LITERAL;
		// System.out.println("\n\nPayload= " + p + "\n\n");
		assertTrue(bag.instructions().contains(p));
		
		// now bag "FOUR"
		assertTrue(codeStorage.containsBag("FOUR"));		
		bag= codeStorage.getBag("FOUR");
		// "FOUR" is a startup bag
		startupRefList.add(bag);

		p= new Payload();
		p.source= 5;
		p.destination= 6;
		p.data= null;
		p.repeats= 3;
		p.type= Payload.PayloadType.INSTRUCTION;
		// System.out.println("\n\nPayload= " + p + "\n\n");
		assertTrue(bag.instructions().contains(p));

		p= new Payload();
		p.source= -1;
		p.destination= 7;
		p.data= new Data(1);
		p.repeats= -1;
		p.type= Payload.PayloadType.LITERAL;
		// System.out.println("\n\nPayload= " + p + "\n\n");
		assertTrue(bag.instructions().contains(p));

		
		// now bag "FIVE" - coming from the included file
		assertTrue(codeStorage.containsBag("FIVE"));		
		bag= codeStorage.getBag("FIVE");
		// "FIVE" is a startup bag - it was loaded first, add it to the beginning of the list
		startupRefList.addFirst(bag);

		p= new Payload();
		p.source= 9;
		p.destination= 10;
		p.data= null;
		p.repeats= 1;
		p.type= Payload.PayloadType.INSTRUCTION;
		// System.out.println("\n\nPayload= " + p + "\n\n");
		assertTrue(bag.instructions().contains(p));

		p= new Payload();
		p.source= -1;
		p.destination= 12;
		p.data= new Data("fleet");
		p.repeats= -1;
		p.type= Payload.PayloadType.LITERAL;
		// System.out.println("\n\nPayload= " + p + "\n\n");
		assertTrue(bag.instructions().contains(p));
		
		
		// check startup bags
		// extract the list of startup bags from code storage
		LinkedList<CodeBag> startupSimList= new LinkedList<CodeBag>();		
		Iterator<CodeBag> cbi= codeStorage.startupIterator();
		while(cbi.hasNext()) {
			CodeBag cb= cbi.next();
			// System.out.println(cb);
			startupSimList.add(cb);
		}
		// the reference list and the assembled list must be the same
		assertTrue(startupRefList.equals(startupSimList));
		
	}		
	
} // class ProgramParserTest
