/*
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */

/*
 * Created on Mar 24, 2005
 *
 */
package com.sunlabs.archsim.tests.branch;

import com.sunlabs.archsim.base.hierarchy.CompositeEntity;
import com.sunlabs.archsim.gasp.State;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;

/**
 * @author ib27688
 */
public class BranchTestFactory {

	private BranchTestFactory() {}
	
	public static CompositeEntity please(String name) {
		CompositeEntity test= new CompositeEntity(name);
		
		// make the source
		BranchTestSource src= new BranchTestSource(BranchTestGlobal.SOURCE, 0, 50, test);
		src.setDelay(BranchTestGlobal.SOURCE_DELAY);
		
		// make the output terminal, set delays, attach it to the source
		OutputTerminal srcOut= new OutputTerminal(BranchTestGlobal.SRC_OUT, test);
		srcOut.setDelayFromActionToState(BranchTestGlobal.ACTION_TO_STATE_DELAY);
		srcOut.setDelayFromStateToAction(BranchTestGlobal.STATE_TO_ACTION_DELAY);
		// now attach
		src.attach(srcOut);
		
		// make the input terminal, set delays, attach it to the source
		InputTerminal srcAckOut= new InputTerminal(BranchTestGlobal.SRC_ACK_OUT, test);
		srcAckOut.setDelayFromActionToState(BranchTestGlobal.ACTION_TO_STATE_DELAY);
		srcAckOut.setDelayFromStateToAction(BranchTestGlobal.STATE_TO_ACTION_DELAY);
		// now attach
		src.attach(srcAckOut);
		
		// make a state: data from src to branch
		State srcDataState= new State(BranchTestGlobal.SOURCE_DATA_STATE, test);		

		// make a state: ack from src to branch
		State srcAckDataState= new State(BranchTestGlobal.ACK_SOURCE_DATA_STATE, test);		

		// Make the branch
		TestBranch branch= new TestBranch(BranchTestGlobal.BRANCH,
								BranchTestGlobal.NUM_BRANCHES);
		
		// attach the terminals to the states
		srcDataState.attach(srcOut);
		srcDataState.attach(branch.getIn());
		
		srcAckDataState.attach(branch.getAckIn());
		srcAckDataState.attach(srcAckOut);
		
		// now make the sinks
		BranchTestSink[] sink= new BranchTestSink[BranchTestGlobal.NUM_BRANCHES];
		State branchDataState[] = new State[BranchTestGlobal.NUM_BRANCHES];
		State ackBranchDataState[] = new State[BranchTestGlobal.NUM_BRANCHES];
		InputTerminal[] sinkIn= new InputTerminal[BranchTestGlobal.NUM_BRANCHES]; 
		OutputTerminal[] sinkAckIn= new OutputTerminal[BranchTestGlobal.NUM_BRANCHES]; 
		
		for (int i=0; i< BranchTestGlobal.NUM_BRANCHES; i++ ) {
			// make the sink
			sink[i]= new BranchTestSink(BranchTestGlobal.SINK+i, test);
			sink[i].setDelay(BranchTestGlobal.SINK_DELAY);
		
			// input to the sink
			sinkIn[i]= new InputTerminal(BranchTestGlobal.SINK_IN+i);
			sinkIn[i].setDelayFromActionToState(BranchTestGlobal.ACTION_TO_STATE_DELAY);
			sinkIn[i].setDelayFromStateToAction(BranchTestGlobal.STATE_TO_ACTION_DELAY);
			// attach
			sink[i].attach(sinkIn[i]);
			
			// output from the sink
			sinkAckIn[i]= new OutputTerminal(BranchTestGlobal.SINK_ACK_IN+i, test);
			sinkAckIn[i].setDelayFromActionToState(BranchTestGlobal.ACTION_TO_STATE_DELAY);
			sinkAckIn[i].setDelayFromStateToAction(BranchTestGlobal.STATE_TO_ACTION_DELAY);
			//attach
			sink[i].attach(sinkAckIn[i]);
			
			// make the state: data from branch to sink
			branchDataState[i]= new State(BranchTestGlobal.BRANCH_DATA_STATE+i, test);
			// attach terminals
			branchDataState[i].attach(branch.getOut(i));
			branchDataState[i].attach(sinkIn[i]);
			
			// make the state: ack from sink to branch
			ackBranchDataState[i]= new State(BranchTestGlobal.ACK_BRANCH_DATA_STATE+i, test);
			// attach the terminals
			ackBranchDataState[i].attach(sinkAckIn[i]);
			ackBranchDataState[i].attach(branch.getAckOut(i));

		}
		
		// initialize active states
		srcAckDataState.setActive(null);
		for (int i= 0; i< BranchTestGlobal.NUM_BRANCHES; i++) {
			ackBranchDataState[i].setActive(null);
		}
		 
		return test;
	} // please
	
}
